{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# 自定义的文本编辑器\n",
    "\n",
    "最终的效果图：\n",
    "\n",
    "![自定义的文本编辑器](https://upload-images.jianshu.io/upload_images/1114626-d710858e785226b6.png?imageMogr2/auto-orient/strip%7CimageView2/2/w/1240)\n",
    "\n",
    "代码：\n",
    "\n",
    "```python\n",
    "from tkinter import filedialog, Toplevel, Menu, Text, Tk, messagebox\n",
    "from tkinter import ttk, PhotoImage, IntVar, StringVar, BooleanVar\n",
    "from pathlib import Path\n",
    "\n",
    "\n",
    "class WindowMeta(Tk):\n",
    "    '''Create a text editor.'''\n",
    "\n",
    "    def __init__(self):\n",
    "        super().__init__()\n",
    "        self.geometry('350x350')\n",
    "        self.show_line_number = IntVar()\n",
    "        self.show_line_number.set(1)\n",
    "        self.create_widgets()\n",
    "        self.layout()\n",
    "        self.content_text.bind('<Any-KeyPress>', self.on_content_changed)\n",
    "        self.content_text.tag_configure('active_line', background='ivory2')\n",
    "        self.protocol('WM_DELETE_WINDOW', self.exit_editor)\n",
    "\n",
    "    def get_line_numbers(self):\n",
    "        output = ''\n",
    "        if self.show_line_number.get():\n",
    "            row, _ = self.content_text.index(\"end\").split('.')\n",
    "            for k in range(1, int(row)):\n",
    "                output += str(k) + '\\n'\n",
    "        return output\n",
    "\n",
    "    def update_line_numbers(self, event=None):\n",
    "        line_numbers = self.get_line_numbers()\n",
    "        self.line_number_bar.config(state='normal')\n",
    "        self.line_number_bar.delete('1.0', 'end')\n",
    "        self.line_number_bar.insert('1.0', line_numbers)\n",
    "        self.line_number_bar.config(state='disabled')\n",
    "\n",
    "    def _create_content_text(self):\n",
    "        '''Create a scrollable text box.'''\n",
    "        self.content_text = Text(wrap='word', undo=1)\n",
    "        self.scroll_bar = ttk.Scrollbar(self.content_text)\n",
    "        self.content_text.configure(yscrollcommand=self.scroll_bar.set)\n",
    "        self.scroll_bar.config(command=self.content_text.yview)\n",
    "\n",
    "    def _create_shortbar(self):\n",
    "        style = ttk.Style()\n",
    "        style.configure('TFrame', background='light sea green')\n",
    "        self.shortcut_bar = ttk.Frame(height=25)\n",
    "\n",
    "    def create_widgets(self):\n",
    "        self._create_shortbar()\n",
    "        self.line_number_bar = Text(width=4, padx=3, takefocus=0,  border=0,\n",
    "                                    background='khaki', state='disabled',  wrap='none')\n",
    "        self._create_content_text()\n",
    "        self.cursor_info_bar = ttk.Label(\n",
    "            self.content_text, text='Row: 0 | Column: 0')\n",
    "\n",
    "    def update_cursor_info_bar(self, event=None):\n",
    "        row, col = self.content_text.index('insert').split('.')\n",
    "        infotext = f\"Row: {row} | Column: {int(col)+1}\"\n",
    "        self.cursor_info_bar.config(text=infotext)\n",
    "\n",
    "    def on_content_changed(self, event=None):\n",
    "        self.update_line_numbers()\n",
    "        self.update_cursor_info_bar()\n",
    "\n",
    "    def layout(self):\n",
    "        '''Initialize the style that sets the Window.'''\n",
    "        self.shortcut_bar.pack(expand='no', fill='x')\n",
    "        self.line_number_bar.pack(side='left',  fill='y')\n",
    "        self.content_text.pack(expand='yes', fill='both')\n",
    "        self.scroll_bar.pack(side='right', fill='y')\n",
    "        self.cursor_info_bar.pack(\n",
    "            expand='no', fill=None, side='right', anchor='se')\n",
    "\n",
    "    def exit_editor(self):\n",
    "        if messagebox.askokcancel(\"Quit?\", \"Really quit?\"):\n",
    "            self.destroy()\n",
    "\n",
    "\n",
    "class Window(WindowMeta):\n",
    "\n",
    "    file_name_params = {\n",
    "        'defaultextension': \".txt\",\n",
    "        'filetypes': [(\"Text Documents\", \"*.txt\"),\n",
    "                      (\"Images\", \"*.jpg *.gif *.png\"),\n",
    "                      (\"All Files\", \"*.*\")]\n",
    "    }\n",
    "\n",
    "    def __init__(self, program_name=\"Notebook Editor\"):\n",
    "        super().__init__()\n",
    "        self.program_name = program_name\n",
    "        self.temp_file_name = StringVar()\n",
    "        self.temp_file_name.trace_add('write', self.update_title)\n",
    "        self.file_name = StringVar()\n",
    "        self.update_title()\n",
    "        self.bind_content_text()\n",
    "\n",
    "    def update_title(self, *args):\n",
    "        file_name = self.temp_file_name.get()\n",
    "        if file_name:\n",
    "            file_name = Path(file_name)\n",
    "            base_name = file_name.parts[-1]\n",
    "            self.title(f\"{base_name}-{self.program_name}\")\n",
    "        else:\n",
    "            self.title(self.program_name)\n",
    "\n",
    "    def display_about_messagebox(self, event=None):\n",
    "        text = \"\\nTkinter GUI Application\\n Development Blueprints\"\n",
    "        messagebox.showinfo(\"About\", f\"{self.program_name}{text}\")\n",
    "\n",
    "    def layout(self):\n",
    "        '''Initialize the style that sets the Window.'''\n",
    "        self.shortcut_bar.pack(expand='no', fill='x')\n",
    "        icons = ('new_file', 'open_file', 'save', 'cut', 'copy', 'paste',\n",
    "                 'undo', 'redo', 'find_text')\n",
    "        for icon in icons:\n",
    "            tool_bar_icon = PhotoImage(file=f'icons/{icon}.gif')\n",
    "            cmd = eval(f'self.{icon}')\n",
    "            tool_bar = ttk.Button(\n",
    "                self.shortcut_bar, image=tool_bar_icon, command=cmd)\n",
    "            tool_bar.image = tool_bar_icon\n",
    "            tool_bar.pack(side='left')\n",
    "        self.line_number_bar.pack(side='left',  fill='y')\n",
    "        self.content_text.pack(expand='yes', fill='both')\n",
    "        self.scroll_bar.pack(side='right', fill='y')\n",
    "        self.cursor_info_bar.pack(\n",
    "            expand='no', fill=None, side='right', anchor='se')\n",
    "\n",
    "    def new_file(self, event=None):\n",
    "        self.temp_file_name.set('Untitled')\n",
    "        self.file_name.set('')\n",
    "        self.content_text.delete(1.0, 'end')\n",
    "        self.on_content_changed()\n",
    "\n",
    "    def open_file(self):\n",
    "        input_file_name = filedialog.askopenfilename(**self.file_name_params)\n",
    "        self.temp_file_name.set(input_file_name)\n",
    "        if input_file_name:\n",
    "            self.file_name.set(input_file_name)\n",
    "            self.content_text.delete(1.0, 'end')\n",
    "            with open(input_file_name, encoding='utf-8') as _file:\n",
    "                self.content_text.insert(1.0, _file.read())\n",
    "            self.on_content_changed()\n",
    "\n",
    "    def write_to_file(self, file_name):\n",
    "        try:\n",
    "            content = self.content_text.get(1.0, 'end')\n",
    "            with open(file_name, 'w', encoding='utf-8') as the_file:\n",
    "                the_file.write(content)\n",
    "        except IOError:\n",
    "            messagebox.showwarning(\"Save\", \"Could not save the file.\")\n",
    "\n",
    "    def save_as(self, event=None):\n",
    "        input_file_name = filedialog.asksaveasfilename(**self.file_name_params)\n",
    "        self.temp_file_name.set(input_file_name)\n",
    "        if input_file_name:\n",
    "            self.write_to_file(input_file_name)\n",
    "        return \"break\"\n",
    "\n",
    "    def save(self, event=None):\n",
    "        file_name = self.file_name.get()\n",
    "        if file_name:\n",
    "            self.write_to_file(file_name)\n",
    "        else:\n",
    "            self.save_as()\n",
    "        return \"break\"\n",
    "\n",
    "    def find_text(self, event=None):\n",
    "        search_toplevel = Toplevel(self.master)\n",
    "        search_toplevel.title('Find Text')\n",
    "        search_toplevel.transient(self.master)\n",
    "\n",
    "        ttk.Label(search_toplevel, text=\"Find All:\").grid(\n",
    "            row=0, column=0, sticky='e')\n",
    "\n",
    "        search_entry_widget = ttk.Entry(\n",
    "            search_toplevel, width=25)\n",
    "        search_entry_widget.grid(row=0, column=1, padx=2, pady=2, sticky='we')\n",
    "        search_entry_widget.focus_set()\n",
    "        ignore_case_value = IntVar()\n",
    "        ttk.Checkbutton(search_toplevel, text='Ignore Case', variable=ignore_case_value).grid(\n",
    "            row=1, column=1, sticky='e', padx=2, pady=2)\n",
    "        button = ttk.Button(search_toplevel, text=\"Find All\", underline=0,\n",
    "                            command=lambda: self.search_output(search_entry_widget.get(),\n",
    "                                                               ignore_case_value.get(),\n",
    "                                                               search_toplevel, search_entry_widget))\n",
    "        button.grid(row=0,\n",
    "                    column=2, sticky='e' + 'w', padx=2, pady=2)\n",
    "\n",
    "        def close_search_window():\n",
    "            self.content_text.tag_remove('match', '1.0', 'end')\n",
    "            search_toplevel.destroy()\n",
    "        search_toplevel.protocol('WM_DELETE_WINDOW', close_search_window)\n",
    "        return \"break\"\n",
    "\n",
    "    def search_output(self, needle, if_ignore_case, search_toplevel, search_box):\n",
    "        self.content_text.tag_remove('match', '1.0', 'end')\n",
    "        matches_found = 0\n",
    "        if needle:\n",
    "            start_pos = '1.0'\n",
    "            while True:\n",
    "                start_pos = self.content_text.search(needle, start_pos,\n",
    "                                                     nocase=if_ignore_case, stopindex='end')\n",
    "                if not start_pos:\n",
    "                    break\n",
    "                end_pos = f'{start_pos}+{len(needle)}c'\n",
    "                self.content_text.tag_add('match', start_pos, end_pos)\n",
    "                matches_found += 1\n",
    "                start_pos = end_pos\n",
    "            self.content_text.tag_config(\n",
    "                'match', foreground='red', background='yellow')\n",
    "        search_box.focus_set()\n",
    "        search_toplevel.title(f'{matches_found} matches found')\n",
    "\n",
    "    def cut(self):\n",
    "        self.content_text.event_generate(\"<<Cut>>\")\n",
    "        self.on_content_changed()\n",
    "        return \"break\"\n",
    "\n",
    "    def copy(self):\n",
    "        self.content_text.event_generate(\"<<Copy>>\")\n",
    "        return \"break\"\n",
    "\n",
    "    def paste(self):\n",
    "        self.content_text.event_generate(\"<<Paste>>\")\n",
    "        self.on_content_changed()\n",
    "        return \"break\"\n",
    "\n",
    "    def undo(self):\n",
    "        self.content_text.event_generate(\"<<Undo>>\")\n",
    "        self.on_content_changed()\n",
    "        return \"break\"\n",
    "\n",
    "    def redo(self, event=None):\n",
    "        self.content_text.event_generate(\"<<Redo>>\")\n",
    "        self.on_content_changed()\n",
    "        return 'break'\n",
    "\n",
    "    def display_help_messagebox(self, event=None):\n",
    "        messagebox.showinfo(\n",
    "            \"Help\", \"Help Book: \\nTkinter GUI Application\\n Development Blueprints\",\n",
    "            icon='question')\n",
    "\n",
    "    def select_all(self, event=None):\n",
    "        self.content_text.tag_add('sel', '1.0', 'end')\n",
    "        return \"break\"\n",
    "\n",
    "    def bind_content_text(self):\n",
    "        self.content_text.bind('<KeyPress-F1>', self.display_help_messagebox)\n",
    "        self.content_text.bind('<Control-N>', self.new_file)\n",
    "        self.content_text.bind('<Control-n>', self.new_file)\n",
    "        self.content_text.bind('<Control-O>', self.open_file)\n",
    "        self.content_text.bind('<Control-o>', self.open_file)\n",
    "        self.content_text.bind('<Control-S>', self.save)\n",
    "        self.content_text.bind('<Control-s>', self.save)\n",
    "        self.content_text.bind('<Control-f>', self.find_text)\n",
    "        self.content_text.bind('<Control-F>', self.find_text)\n",
    "        self.content_text.bind('<Control-A>', self.select_all)\n",
    "        self.content_text.bind('<Control-a>', self.select_all)\n",
    "        self.content_text.bind('<Control-y>', self.redo)\n",
    "        self.content_text.bind('<Control-Y>', self.redo)\n",
    "\n",
    "\n",
    "class Popup:\n",
    "    def __init__(self, master):\n",
    "        self.master = master\n",
    "        self.generate_menu()\n",
    "        # bind right mouse click to show pop up and set focus to text widget on launch\n",
    "        self.master.content_text.bind('<Button-3>', self.show_popup_menu)\n",
    "        self.master.content_text.focus_set()\n",
    "\n",
    "    def generate_menu(self):\n",
    "        # set up the pop-up menu\n",
    "        self.popup_menu = Menu(self.master.content_text, tearoff=0)\n",
    "        for m in ('cut', 'copy', 'paste', 'undo', 'redo'):\n",
    "            cmd = eval(f'self.master.{m}')\n",
    "            self.popup_menu.add_command(label=m, compound='left', command=cmd)\n",
    "        self.popup_menu.add_separator()\n",
    "        self.popup_menu.add_command(\n",
    "            label='Select All', underline=7, command=self.master.select_all)\n",
    "\n",
    "    def show_popup_menu(self, event):\n",
    "        self.popup_menu.tk_popup(event.x_root, event.y_root)\n",
    "\n",
    "\n",
    "class File(Menu):\n",
    "    def __init__(self, master, action, **kw):\n",
    "        super().__init__(master, **kw)\n",
    "        self.master = master\n",
    "        self.action = action\n",
    "        self.generate_menu()\n",
    "\n",
    "    def _load_icon(self):\n",
    "        self._new_file_icon = PhotoImage(file='icons/new_file.gif')\n",
    "        self._open_file_icon = PhotoImage(file='icons/open_file.gif')\n",
    "        self._save_file_icon = PhotoImage(file='icons/save.gif')\n",
    "\n",
    "    def _create_menu(self):\n",
    "        self.add_command(label='New', accelerator='Ctrl+N', command=self.action.new_file,\n",
    "                         compound='left', image=self._new_file_icon, underline=0)\n",
    "        self.add_command(label='Open', accelerator='Ctrl+O', command=self.action.open_file,\n",
    "                         compound='left', image=self._open_file_icon, underline=0)\n",
    "        self.add_command(label='Save', accelerator='Ctrl+S', command=self.action.save,\n",
    "                         compound='left', image=self._save_file_icon, underline=0)\n",
    "        self.add_command(\n",
    "            label='Save as', accelerator='Shift+Ctrl+S', command=self.action.save_as)\n",
    "        self.add_separator()\n",
    "        self.add_command(label='Exit', accelerator='Alt+F4',\n",
    "                         command=self.action.exit_editor)\n",
    "\n",
    "    def generate_menu(self):\n",
    "        self._load_icon()\n",
    "        self._create_menu()\n",
    "\n",
    "\n",
    "class Edit(Menu):\n",
    "    def __init__(self, master, action, **kw):\n",
    "        super().__init__(master, **kw)\n",
    "        self.action = action\n",
    "        self.generate_menu()\n",
    "\n",
    "    def _load_icon(self):\n",
    "        self._cut_icon = PhotoImage(file='icons/cut.gif')\n",
    "        self._copy_icon = PhotoImage(file='icons/copy.gif')\n",
    "        self._paste_icon = PhotoImage(file='icons/paste.gif')\n",
    "        self._undo_icon = PhotoImage(file='icons/undo.gif')\n",
    "        self._redo_icon = PhotoImage(file='icons/redo.gif')\n",
    "\n",
    "    def _create_menu(self):\n",
    "        self.add_command(label='Undo', accelerator='Ctrl+Z',\n",
    "                         compound='left', image=self._undo_icon, command=self.action.undo)\n",
    "        self.add_command(label='Redo', accelerator='Ctrl+Y',\n",
    "                         compound='left', image=self._redo_icon, command=self.action.redo)\n",
    "        self.add_separator()\n",
    "        self.add_command(label='Cut', accelerator='Ctrl+X',\n",
    "                         compound='left', image=self._cut_icon, command=self.action.cut)\n",
    "        self.add_command(label='Copy', accelerator='Ctrl+C',\n",
    "                         compound='left', image=self._copy_icon, command=self.action.copy)\n",
    "        self.add_command(label='Paste', accelerator='Ctrl+V',\n",
    "                         compound='left', image=self._paste_icon, command=self.action.paste)\n",
    "        self.add_separator()\n",
    "        self.add_command(\n",
    "            label='Find', underline=0, accelerator='Ctrl+F', command=self.action.find_text)\n",
    "        self.add_separator()\n",
    "        self.add_command(label='Select All', underline=7,\n",
    "                         accelerator='Ctrl+A', command=self.action.select_all)\n",
    "\n",
    "    def generate_menu(self):\n",
    "        self._load_icon()\n",
    "        self._create_menu()\n",
    "\n",
    "\n",
    "class View(Menu):\n",
    "    color_schemes = {\n",
    "        'Default': '#000000.#FFFFFF',\n",
    "        'Greygarious': '#83406A.#D1D4D1',\n",
    "        'Aquamarine': '#5B8340.#D1E7E0',\n",
    "        'Bold Beige': '#4B4620.#FFF0E1',\n",
    "        'Cobalt Blue': '#ffffBB.#3333aa',\n",
    "        'Olive Green': '#D1E7E0.#5B8340',\n",
    "        'Night Mode': '#FFFFFF.#000000',\n",
    "    }\n",
    "\n",
    "    def __init__(self, master, action, **kw):\n",
    "        super().__init__(master, **kw)\n",
    "        self.master = master\n",
    "        self.action = action\n",
    "        self.show_cursor_info = IntVar()\n",
    "        self.show_cursor_info.set(1)\n",
    "        self.generate_menu()\n",
    "\n",
    "    def highlight_line(self, interval=100):\n",
    "        self.action.content_text.tag_remove(\"active_line\", 1.0, \"end\")\n",
    "        self.action.content_text.tag_add(\n",
    "            \"active_line\", \"inset linestart\", \"insert lineend+1c\")\n",
    "        self.action.content_text.after(interval, self.toggle_highlight)\n",
    "\n",
    "    def undo_highlight(self):\n",
    "        self.action.content_text.tag_remove(\"active_line\", 1.0, \"end\")\n",
    "\n",
    "    def toggle_highlight(self, event=None):\n",
    "        if self.to_highlight_line.get():\n",
    "            self.highlight_line()\n",
    "        else:\n",
    "            self.undo_highlight()\n",
    "\n",
    "    def change_theme(self, event=None):\n",
    "        selected_theme = self.theme_choice.get()\n",
    "        fg_bg_colors = self.color_schemes.get(selected_theme)\n",
    "        foreground_color, background_color = fg_bg_colors.split('.')\n",
    "        self.action.content_text.config(\n",
    "            background=background_color, fg=foreground_color)\n",
    "\n",
    "    def show_cursor_info_bar(self):\n",
    "        show_cursor_info_checked = self.show_cursor_info.get()\n",
    "        if show_cursor_info_checked:\n",
    "            self.action.cursor_info_bar.pack(\n",
    "                expand='no', fill=None, side='right', anchor='se')\n",
    "        else:\n",
    "            self.action.cursor_info_bar.pack_forget()\n",
    "\n",
    "    def generate_menu(self):\n",
    "        self.add_checkbutton(\n",
    "            label='Show Line Number', variable=self.action.show_line_number, command=self.action.update_line_numbers)\n",
    "        self.add_checkbutton(\n",
    "            label='Show Cursor Location at Bottom', variable=self.show_cursor_info, command=self.show_cursor_info_bar)\n",
    "        self.to_highlight_line = BooleanVar()\n",
    "        self.add_checkbutton(label='Highlight Current Line', onvalue=1, offvalue=0,\n",
    "                             variable=self.to_highlight_line, command=self.toggle_highlight)\n",
    "        self.theme_choice = StringVar()\n",
    "        self.theme_choice.set('Default')\n",
    "        self.themes_menu = Menu(self.master, tearoff=0)\n",
    "        for k in sorted(self.color_schemes):\n",
    "            self.themes_menu.add_radiobutton(\n",
    "                label=k, variable=self.theme_choice, command=self.change_theme)\n",
    "        self.add_cascade(label='Themes', menu=self.themes_menu)\n",
    "\n",
    "\n",
    "class About(Menu):\n",
    "    def __init__(self, master, action, **kw):\n",
    "        super().__init__(master, **kw)\n",
    "        self.action = action\n",
    "        self.generate_menu()\n",
    "\n",
    "    def generate_menu(self):\n",
    "        self.add_command(\n",
    "            label='Help', command=self.action.display_help_messagebox)\n",
    "        self.add_command(\n",
    "            label='About', command=self.action.display_about_messagebox)\n",
    "\n",
    "\n",
    "def test():\n",
    "    root = Window()\n",
    "    menu_bar = Menu(root)\n",
    "    root['menu'] = menu_bar\n",
    "\n",
    "    file_menu = File(menu_bar, root, tearoff=0)\n",
    "    menu_bar.add_cascade(label='File', menu=file_menu)\n",
    "\n",
    "    edit_menu = Edit(menu_bar, root, tearoff=0)\n",
    "    menu_bar.add_cascade(label='Edit', menu=edit_menu)\n",
    "\n",
    "    view_menu = View(menu_bar, root, tearoff=0)\n",
    "    menu_bar.add_cascade(label='View', menu=view_menu)\n",
    "\n",
    "    about_menu = About(menu_bar, root, tearoff=0)\n",
    "    menu_bar.add_cascade(label='About',  menu=about_menu)\n",
    "\n",
    "    Popup(root)\n",
    "    root.mainloop()\n",
    "\n",
    "\n",
    "if __name__ == '__main__':\n",
    "    test()\n",
    "```"
   ]
  }
 ],
 "metadata": {
  "language_info": {
   "name": "python"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
